/*
 *  robotfindskitten: A Zen Simulation -- Java
 *
 *  Copyright (C) 2001 Brian Enigma <enigma@netninja.com>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or EXISTANCE OF KITTEN.  See the GNU General
 *   Public License for more details.
 *
 *   http://www.gnu.org/copyleft/gpl.html
 */
package org.zensimulation.robotfindskitten;

import java.util.*;
import java.awt.Point;

/**
 *  Defines the datastructures and operations relating to a game board.
 *  This includes creating it, populating it with random items,
 *  moving about it, and detecting collisions.
 *
 *@author     Brian Enigma
 *@created    July 2001
 *@version    $Id$
 */
public class Board
{

    /** Width, in grid squares, of the board */
    private int width;
    /** Height, in grid squares, of the board */
    private int height;
    /** Board contents.  Hashtable maps Point(x,y) => BoardItem() */
    private Hashtable board;
    /** Simple vector of strings, one entry for each message that has
        been populated on the board.  I really wanted this to be a
        TreeSet for quick indexing, but unfortunately that was added
        to Java 1.2 (which many/most browsers do not support). */
    private Vector seenMessages;
    /** The position of the robot on the board.  Unless alternate-world
        physics are used, this should be a point that does not exist in
        the this.board Hashtable. */
    private Point robotPosition;
    /** The last message that the user received (from bumping into an item). */
    private String lastMessage;
    /** The symbol associated with the last thing the user bumped into.  THis
        is really only useful for displaying the kitten "win" animation. */
    private String lastSymbol;
    /** Return value from move, indicates user ran into wall.  Robot
        position has not changed.  */
    public final static int CANT_MOVE = 0;
    /** Return value from move, indicates user moved into blank square.  
        Robot position has changed.  */
    public final static int MOVED = 1;
    /** Return value from move, indicates user hit an object.  Robot position
        has not changed.  lastMessage and lastSymbol members have been
        updated to reflect object. */
    public final static int HIT_OBJECT = 2;

    /** Parameter to move, move up. */
    public final static int DIR_UP = 0;
    /** Parameter to move, move right. */
    public final static int DIR_RIGHT = 1;
    /** Parameter to move, move down. */
    public final static int DIR_DOWN = 2;
    /** Parameter to move, move left. */
    public final static int DIR_LEFT = 3;

    /**
     *  Message string for the object that "is kitten."  Everything is
     *  "not kitten" and should be displayed to user.
     */
    public final static String KITTEN_STR = "KITTEN";

    /**
     *  This is our singleton pseudorandom number generator.  It is seeded 
     *  from the system clock, so if multiple Random objects are initialized
     *  very quickly (i.e. from multiple calls to the same function that
     *  utilizes random numbers), they chould yield the same "random"
     *  number each time.
     */
    public static Random r = new Random();


    /**
     *  Construct a default-sized board, with the given number of objects.
     *
     *@param  objectCount  Number of objects to place on the board.
     */
    public Board(int objectCount)
    {
        this.width = 79;
        this.height = 20;
        this.board = new Hashtable();
        this.seenMessages = new Vector();
        this.lastMessage = new String();
        this.lastSymbol = null;
        randomize(objectCount);
    }


    /**
     *  Construct a board object with a width, height, and object density.
     *
     *@param  objectCount  Number of objects to place on the board.
     *@param  width        Width, in grid squares, of the board.
     *@param  height       Height, in grid squares, of the board.
     */
    public Board(int objectCount, int width, int height)
    {
        this.width = width;
        this.height = height;
        this.board = new Hashtable();
        this.seenMessages = new Vector();
        this.lastMessage = null;
        this.lastSymbol = new String();
        // Sanity check
        width = Math.min(width, 79);
        width = Math.max(2, width);
        height = Math.min(height, 20);
        height = Math.max(2, height);
        // Populate board
        randomize(objectCount);
    }


    /**
     *  Private parameterless constructor--you can't do this.
     */
    private Board()
    {
    }


    /**
     *  Returns the robot X coordinate
     *
     *@return    The robot X coordinate
     */
    public int getRobotX()
    {
        return robotPosition.x;
    }


    /**
     *  Returns the robot Y coordinate
     *
     *@return    The robot Y coordinate
     */
    public int getRobotY()
    {
        return robotPosition.y;
    }


    /**
     *  Gets the last message received from bumping into an object
     *
     *@return    The string (which through massive amounts of careful
     *           planning could be a message, could be an empty string, or
     *           could be null.
     */
    public String getLastMessage()
    {
        return this.lastMessage;
    }


    /**
     *  Gets the last symbol that robot bumped into.
     *
     *@return    The string (which, again, though careful planning and
     *           coordination could be a real string, an empty string,
     *           or null)
     */
    public String getLastSymbol()
    {
        return this.lastSymbol;
    }


    /**
     *  Gets the symbol string located at the given (x,y) position.  This
     *  is used by the UI to repaint the board.
     *
     *@param  x  The x coordinate
     *@param  y  The y coordinate
     *@return    The symbol string or null, if no symbol
     */
    public String getBoardSymbol(int x, int y)
    {
        Point p = new Point(x, y);
        BoardItem item = (BoardItem) (this.board.get(p));
        if (item == null)
        {
            return null;
        }
        return item.symbol;
    }


    /**
     *  Move the robot in the given direction, return information about
     *  the move.
     *
     *@param  direction  A move constant (up,down,right,left)
     *@return            A disposition (moved,hitwall,hitobject)
     */
    public int move(int direction)
    {
        Point p = new Point(robotPosition);
        int deltaX = 0;
        int deltaY = 0;
        this.lastMessage = "";
        this.lastSymbol = null;
        switch (direction)
        {
            case DIR_UP:
                deltaY = -1;
                break;
            case DIR_DOWN:
                deltaY = 1;
                break;
            case DIR_RIGHT:
                deltaX = 1;
                break;
            case DIR_LEFT:
                deltaX = -1;
                break;
        }
        p.translate(deltaX, deltaY);
        if ((p.x < 0) || (p.y < 0) || (p.x >= width) || (p.y >= height))
        {
            return CANT_MOVE;
        }
        if (board.containsKey(p))
        {
            this.lastMessage = ((BoardItem) board.get(p)).message;
            this.lastSymbol = ((BoardItem) board.get(p)).symbol;
            return HIT_OBJECT;
        }
        robotPosition = p;
        return MOVED;
    }


    /**
     *  Dump the board to a debug string.  This is never used during normal
     *  program operation, only during debugging.
     *
     *@return    Description of the Returned Value
     */
    public String toString()
    {
        StringBuffer b = new StringBuffer();
        for (Enumeration i = board.keys(); i.hasMoreElements(); )
        {
            Point key = (Point) (i.nextElement());
            BoardItem value = (BoardItem) board.get(key);
            b.append("(");
            b.append(key.x);
            b.append(",");
            b.append(key.y);
            b.append(") ");
            b.append(value.toString());
            b.append("\n");
        }
        return b.toString();
    }


    /**
     *  Hooooyeah, we like multiple linear searches through linear
     *  datastructures. "We'll optimize in version 2.0"  Really, this
     *  was supposed to be a very efficient lookup/insert into a very
     *  keen binary-tree driven Set object.  Unfortunately, this was
     *  only useful in Java 1.2, not in browser java.
     *
     *@param  messageNumber  The message number that interests us.
     *@return                Whether the message has been seen.
     */
    private boolean isDuplicateMessage(Integer messageNumber)
    {
        int max = this.seenMessages.size();
        for (int i = 0; i < max; i++)
        {
            if (((Integer) (this.seenMessages.elementAt(i))).equals(messageNumber))
            {
                return false;
            }
        }
        return true;
    }


    /**
     *  Randomize the board, using the objects member variables for size and
     *  object density.  Place the objects, then place the kitten (so as not
     *  to occupy the same spacetime coordinates as anything else), then
     *  place the robot (so as not to occupy the same spacetime coordinates
     *  as the items or kitten).  Maybe we could come up with with a new
     *  Buckaroo Banzai version that allows two items to occupy the same
     *  spot?  Maybe we could come up with a Schrodinger version in which
     *  the kitty is simultaneously alive and dead?  But I digress...
     *
     *  Please make sure that the object count is reasonable.  If the
     *  items are too dense, it won't work...it will simply cycle through
     *  the loops, trying to place items in a completely full board.  Also
     *  if the items are way too dense, there is a very good chance that
     *  the kitten will be blocked on all sides by objects.
     *
     *@param  objectCount  Number of non-kitten, non-robot objects to place 
     *                     on the board
     */
    private void randomize(int objectCount)
    {
        int x;
        int y;
        int count;
        Point p;
        BoardItem bi;
        Integer key;
        //Random r = new Random();
        // Sanity check
        if (objectCount >= (width * height) - 1)
        {
            objectCount = (width * height) - 1;
        }
        if (objectCount < 1)
        {
            objectCount = 1;
        }
        for (count = 0; count < objectCount; count++)
        {
            // This could get ugly if the board is very dense...it
            // repeats, generating a random point until unique
            do
            {
                p = new Point(nextInt(width), nextInt(height));
            } while (board.containsKey(p));
            // Repeats, generating a message until unique
            bi = new BoardItem();
            while (!isDuplicateMessage(bi.messageNumber))
            {
                bi.newMessage();
            }
            seenMessages.addElement(bi.messageNumber);
            board.put(p, bi);
        }
        // Repeats, generating a kitten position until unique
        do
        {
            p = new Point(nextInt(width), nextInt(height));
        } while (board.containsKey(p));
        bi = new BoardItem();
        bi.message = KITTEN_STR;
        board.put(p, bi);
        // Repeats, generating a robot position until unique
        do
        {
            p = new Point(nextInt(width), nextInt(height));
        } while (board.containsKey(p));
        robotPosition = p;
    }


    /**
     *  Get the next integer from the pseudorandom number generator.  This
     *  very function is defined in Java 1.2 in the Random class, but is
     *  not in previous versions, so I had to write it myself here.
     *
     *@param  i  Pick a random number 0 <= rand < i
     *@return    The random value
     */
    private int nextInt(int i)
    {
        // The pseudorandom number generator on my java VM (Linux Netscape)
        // does not seem to have an even Modulus distribution for some
        // reason...
        return Math.abs(r.nextInt() % i);
        // This is sloppy and can take an unknown amount of time to return...
        // ...yes, now quanitatively proven.  It takes eons.
        /*
         *  int result;
         *  do
         *  {
         *  result = Math.abs(r.nextInt());
         *  }
         *  while (result >= i);
         *  return result;
         */
    }


    /**
     *  An internal class to hold information about items on the board.
     *  There is one of these for each non-kitten, non-robot item, plus
     *  one for the kitten in which the message is this.KITTEN_STR.
     *
     *@author     Brian Enigma
     *@created    July 2001
     *@version    $Id$
     *
     */
    public class BoardItem
    {
        /** The symbol */
        public String symbol;
        /** The message */
        public String message;
        /** The message number (to help look for duplicates) */
        public Integer messageNumber;
        /** Which symbols are valid to choose from */
        public final static String VALID_SYMBOLS =
            "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890" +
            "-=!@$%^&*()_+;';\\\",./<>?`~";

        /**
         *  Construct the board item, with a random symbol and message.
         *  At the time of construction, the message COULD be a duplicate
         *  of something else that is already on the board.
         */
        public BoardItem()
        {
            //Random r = new Random();
            int pos;
            pos = Math.abs(nextInt(VALID_SYMBOLS.length()));
            this.symbol = VALID_SYMBOLS.substring(pos, pos + 1);
            newMessage();
        }


        /**
         *  Constuct a board item with the given symbol and message.  I
         *  kind of don't think this is even used....?
         *
         *@param  symbol   The symbol
         *@param  message  The message
         */
        public BoardItem(String symbol, String message)
        {
            this.symbol = symbol;
            this.message = message;
        }


        /**
         *  Regenerate a new message.  If the creator of this object decides
         *  that the message is a duplicate, this can be called to generate
         *  a new message.
         */
        public void newMessage()
        {
            int pos;
            pos = Math.abs(nextInt(Messages.messages.length));
            this.message = Messages.messages[pos];
            this.messageNumber = new Integer(pos);
        }


        /**
         *  Return a (debug) representation of this object.
         *
         *@return    Description of point, symbol, and message.
         */
        public String toString()
        {
            StringBuffer b = new StringBuffer();
            b.append(this.symbol);
            b.append(" ");
            b.append(this.message);
            return b.toString();
        }
    }
}

