/*
 *   robotfindskitten: A Zen Simulation -- Java Version
 *   Copyright (C) 2001 Brian Enigma <enigma@netninja.com>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or EXISTANCE OF KITTEN.  See the GNU General
 *   Public License for more details.
 *
 *   http://www.gnu.org/copyleft/gpl.html
 */
package org.zensimulation.robotfindskitten;
import java.awt.*;
import java.awt.event.*;
//import java.awt.image.*;

/**
 *  The robotfindskitten applet class.  The applet can be in one of two
 *  states:
 *
 *  Single threaded state is normal play, in which you see the grid and
 *  can move around, bump into objects, and see messages.  Winning brings
 *  you to the multithreaded state.
 *
 *  Multithreaded state is the winning animation, in which the main thread
 *  sits in the even loop and updates the UI and the background thread
 *  sets which frame of the animation to show and signals to the UI thread
 *  to repaint after each frame.  The end of the animation brings you to
 *  singlethreaded state.
 *
 *@author     Brian Enigma
 *@created    July 2001
 *@version    $Id$
 */
public class RobotfindskittenApplet
     extends java.applet.Applet
     implements KeyListener, Runnable
{

    Board board;
    private int width;
    private int height;
    private int objects;
    private String startupMessage =
        "Click on the grid to let the applet take keyboard focus";
    private String message = startupMessage;
    private String kitten = null;
    private int distance;
    private final static int SQUARE_WIDTH = 10;
    private final static int SQUARE_HEIGHT = 12;
    private final static Color lightCyan = new Color(224, 255, 255);
    private final static Font bigfont = new Font("Monospaced", Font.BOLD, 24);
    private final static String VERSION = "RFK-JAVA 1.1." +
        Integer.toString(Messages.messages.length) + " / Brian Enigma";


    /**
     *  Initialize the applet class, reading parameters from the APPLET PARAM
     *  tags.
     */
    public void init()
    {
        super.init();
        this.width = 0;
        this.height = 0;
        this.objects = 0;
        try {
            this.width = Integer.parseInt(getParameter("w"));
        } catch (Exception e) { }
        try {
            this.height = Integer.parseInt(getParameter("h"));
        } catch (Exception e) { }
        try {
            this.objects = Integer.parseInt(getParameter("o"));
        } catch (Exception e) { }

        if ((this.width < 2) || (this.width > 80))
            this.width = 80;
        if ((this.height < 2) || (this.height > 20))
            this.height = 20;
        if ((this.objects < 2) || (this.objects > 20))
            this.objects = 20;
        this.board = new Board(objects, width, objects);
        setBackground(Color.white);
    }


    /**
     *  Start the applet
     */
    public void start()
    {
        super.start();
        repaint();
        addKeyListener(this);
    }


    /**
     *  Stop the applet
     */
    public void stop()
    {
        removeKeyListener(this);
    }


    /**
     *  Interface for keypressed event...we don't care.
     *
     *@param  e  Description of Parameter
     */
    public void keyPressed(KeyEvent e)
    {
    }


    /**
     *  Interface for key typed event...we don't care because this does not
     *  pick up "special" keys like the arrows.
     *
     *@param  e  Description of Parameter
     */
    public void keyTyped(KeyEvent e)
    {
    }


    /**
     *  Interface for key release event, we read the key and perform an
     *  operation based on it.
     *
     *@param  e  Description of Parameter
     */
    public void keyReleased(KeyEvent e)
    {
        int dir = -1;
        if (this.kitten != null)
        {
            return;
        }
        if (message == startupMessage)
        {
            message = null;
        }
        switch (e.getKeyCode())
        {
            case KeyEvent.VK_UP:
            case KeyEvent.VK_KP_UP:
                dir = Board.DIR_UP;
                break;
            case KeyEvent.VK_DOWN:
            case KeyEvent.VK_KP_DOWN:
                dir = Board.DIR_DOWN;
                break;
            case KeyEvent.VK_LEFT:
            case KeyEvent.VK_KP_LEFT:
                dir = Board.DIR_LEFT;
                break;
            case KeyEvent.VK_RIGHT:
            case KeyEvent.VK_KP_RIGHT:
                dir = Board.DIR_RIGHT;
                break;
        }
        if (dir != -1)
        {
            if (board.move(dir) == Board.HIT_OBJECT)
            {
                this.message = board.getLastMessage();
                if (this.message.equals(Board.KITTEN_STR))
                {
                    this.kitten = board.getLastSymbol();
                    this.message = "";
                    new Thread(this).start();
                }
            }
            repaint();
        }
        switch (e.getKeyChar())
        {
            case 'q':
            case 'Q':
                // "Quit" (reset board)
                this.board = new Board(objects, width, objects);
                this.kitten = null;
            // fall into repaint section...
            case 'r':
            case 'R':
                // Refresh screen
                repaint();
                break;
            case 'd':
            case 'D':
                // Debug
                System.out.println(this.board.toString());
                break;
            case 'v':
            case 'V':
                //Version v = new Version(this);
                //v.show();
                getAppletContext().showStatus(VERSION);
                break;
        }
    }


    /**
     *  Paint the applet.  If we are in the middle of displaying the
     *  winning animation, paint the current frame of the win animation...if
     *  we are playing the simulation, paint the grid and message.
     *
     *@param  g  Graphics context
     */
    public void paint(Graphics g)
    {
        if (kitten != null)
        {
            paintWin(g);
            return;
        }
        Rectangle r = getBounds();
        g.clearRect(r.x, r.y, r.width, r.height);
        int robotX = board.getRobotX();
        int robotY = board.getRobotY();
        /*
         *  Netscape doesn't seem to like BufferedImage/Writable Wraster
         *  BufferedImage bi = new BufferedImage(r.width, r.height,
         *  BufferedImage.TYPE_INT_RGB);
         *  Graphics gr = bi.getGraphics();
         */
        g.clearRect(r.x, r.y, r.width, r.height);
        g.setColor(lightCyan);
        g.drawRect(2, 2, 2 + SQUARE_WIDTH * width,
            2 + SQUARE_HEIGHT * height);
        for (int w = 0; w < width; w++)
        {
            g.drawLine(SQUARE_WIDTH * w + 2,
                2,
                SQUARE_WIDTH * w + 2,
                SQUARE_HEIGHT * height + 2);
        }
        for (int h = 0; h < height; h++)
        {
            g.drawLine(2,
                2 + SQUARE_HEIGHT * h,
                SQUARE_WIDTH * width + 2,
                2 + SQUARE_HEIGHT * h);
        }
        g.setColor(Color.black);
        for (int w = 0; w < width; w++)
        {
            for (int h = 0; h < height; h++)
            {
                paintSquare(g, false, w, h);
            }
        }
        if (message != null)
        {
            g.drawString(message, 4, (height + 1) * SQUARE_HEIGHT);
        }
        g.setColor(Color.blue);
        g.drawString("#", robotX * SQUARE_WIDTH + 4,
            (robotY + 1) * SQUARE_HEIGHT - 1);
        //g.drawImage(bi, 0, 0, null);
        //super.paint(g);
    }


    /**
     *  Paint an individual grid square contents
     *
     *@param  g           Graphics context
     *@param  clearFirst  Whether or not to clear the square (unused)
     *@param  x           X coordinate
     *@param  y           Y coordinage
     */
    public void paintSquare(Graphics g, boolean clearFirst, int x, int y)
    {
        String symbol;
        symbol = this.board.getBoardSymbol(x, y);
        if (symbol != null)
        {
            g.drawString(symbol, x * SQUARE_WIDTH + 4,
                (y + 1) * SQUARE_HEIGHT - 1);
        }
    }


    /**
     *  Paint the current frame of the "win" animation.  Called by paint
     *  when the animation should be shown.
     *
     *@param  g  Description of Parameter
     */
    public void paintWin(Graphics g)
    {
        String message = new String();
        int preSpace = 8 - this.distance - 1;
        int midSpace = this.distance + 1;
        Rectangle r = getBounds();
        g.clearRect(r.x, r.y, r.width, r.height);
        Font oldFont = g.getFont();
        g.setFont(bigfont);
        for (int i = 0; i < preSpace; i++)
        {
            message += " ";
        }
        message += "#";
        for (int i = 0; i < midSpace; i++)
        {
            message += " ";
        }
        message += this.kitten;
        g.drawString(message, 10, 50);
        if (this.distance <= 0)
        {
            g.drawString("You found kitten! Way to go, robot!", 10, 100);
        }

        g.setFont(oldFont);
    }


    /**
     *  Thread for updating state and repainting the win animation
     */
    public void run()
    {
        // Go through each frame of the win animation
        for (this.distance = 8; this.distance >= 0; this.distance--)
        {
            repaint();
            try {
            Thread.sleep(100);
            } catch (Exception ex) { }
        }
        repaint();
        try {
        Thread.sleep(2000);
        } catch (Exception ex) { }
        this.board = new Board(objects, width, objects);
        this.kitten = null;
        repaint();
    }
}
